"use strict";const m_root=self.location.pathname.split("/").slice(0,-1).join("/");importScripts(m_root+"/r2.js");var LINES=80,MAXLINES=Math.round(1.2*LINES),TOOLONG=2*LINES*3;function appendTo(o,s){if(null!==s)for(var e=0;e<s.length;e++){var n=parseInt(s[e].offset);"0"==s[e].size||s[e].size>=TOOLONG||void 0!==o[n]&&s[e].size<=o[n]||(o[n]=parseInt(s[e].size))}}self.onmessage=function(){var s,e,o={};r2.cmdj("fj ",function(o){s=o}),r2.cmdj("aflj",function(o){e=o}),appendTo(o,s),appendTo(o,e),self.postMessage(o)};