#ifndef R_VERSION_H
#define R_VERSION_H 1
#define R_MESON_VERSION "0.64.1"
#define R2_VERSION_MAJOR 5
#define R2_VERSION_MINOR 8
#define R2_VERSION_PATCH 0
#define R2_VERSION_NUMBER 50800
#define R2_VERSION_COMMIT 0
#define R2_VERSION "5.8.0"
#define R2_GITTAP "5.8.0"
#define R2_GITTIP "unknown"
#define R2_BIRTH "2023-01-11__13:09:07"
#endif
